#include <cmath>
#include <msp/strings/format.h>
#include "blend.h"
#include "bloom.h"
#include "misc.h"
#include "renderer.h"
#include "shader.h"
#include "tests.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

Bloom::Bloom(unsigned w, unsigned h):
	blur_shader("bloom_blur.glsl"),
	combine_shader("bloom_combine.glsl"),
	quad(get_fullscreen_quad())
{
	blur_shdata[0].uniform("delta", 1.0f/w, 0.0f);
	blur_shdata[1].uniform("delta", 0.0f, 1.0f/h);

	blur_shdata_common.uniform("source", 0);
	for(unsigned i=0; i<2; ++i)
	{
		tex[i].set_min_filter(NEAREST);
		tex[i].set_wrap(CLAMP_TO_EDGE);
		tex[i].storage(RGB16F, w, h);
		fbo[i].attach(COLOR_ATTACHMENT0, tex[i], 0);
		fbo[i].require_complete();
	}

	combine_shdata.uniform("source", 1);
	combine_shdata.uniform("blurred", 0);

	combine_texturing.attach(0, tex[1]);

	set_radius(2.0f);
	set_strength(0.2f);
}

void Bloom::set_radius(float r)
{
	if(r<=0.0f)
		throw invalid_argument("Bloom::set_radius");

	int size = min(static_cast<int>(r*3.0f), 9);
	blur_shdata_common.uniform("size", size);

	vector<float> factors(size*2+1);
	float sum = 0.0f;
	r = 2*r*r;
	for(int i=-size; i<=size; ++i)
		sum += (factors[size+i] = exp(-i*i/r));
	for(int i=0; i<=size*2; ++i)
		factors[i] /= sum;

	blur_shdata_common.uniform1_array("factors", size*2+1, &factors.front());
}

void Bloom::set_strength(float s)
{
	if(s<0.0f || s>1.0f)
		throw invalid_argument("Bloom::set_strength");
	combine_shdata.uniform("strength", s);
}

void Bloom::render(Renderer &renderer, const Texture2D &src, const Texture2D &)
{
	{
		Renderer::Push push(renderer);
		renderer.set_shader_program(&blur_shader, &blur_shdata_common);
		for(unsigned i=0; i<2; ++i)
		{
			BindRestore bind_fbo(fbo[i]);
			Renderer::Push push2(renderer);
			renderer.set_texture(i ? &tex[0] : &src);
			renderer.add_shader_data(blur_shdata[i]);
			quad.draw(renderer);
		}
	}

	Renderer::Push push(renderer);
	combine_texturing.attach(1, src);
	renderer.set_texturing(&combine_texturing);
	renderer.set_shader_program(&combine_shader, &combine_shdata);
	quad.draw(renderer);
}

} // namespace GL
} // namespace Msp
